"""
.. module stratumgs.web

The web server for stratumgs.
"""

import datetime
import os

import tornado.web
import tornado.websocket
import tornado.httpserver
import tornado.ioloop

import stratumgs.game
import stratumgs.client.server


def init(host, port, debug):
    """
        Initialize the web server.

        :param port: The web server port.
        :type port: int
   """

    template_path = os.path.join(
        os.path.dirname(os.path.realpath(__file__)),
        "assets", "templates")
    static_files_path = os.path.join(
        os.path.dirname(os.path.realpath(__file__)),
        "assets", "web")
    app = tornado.web.Application([
        tornado.web.url(r"/", HomeHandler, name="main"),
        tornado.web.url(r"/games", GamesHandler, name="games"),
        tornado.web.url(r"/games/([^/]+)/configure", ConfigureHandler, name="configure"),
        tornado.web.url(r"/games/([^/]+)/start", StartHandler, name="start"),
        tornado.web.url(r"/games/([^/]+)/view/([\d]+)", ViewHandler, name="view"),
        tornado.web.url(r"/games/([^/]+)/view/([\d]+)/socket", ViewSocketHandler,
                        name="view_socket"),
        tornado.web.url(r"/matches", MatchesHandler, name="matches"),
        tornado.web.url(r"/players", PlayersHandler, name="players"),
        tornado.web.url(r"/assets/(.*)", tornado.web.StaticFileHandler,
                        {"path": static_files_path}, name="static")
    ], template_path=template_path, debug=debug)
    server = tornado.httpserver.HTTPServer(app)
    server.listen(port, address=host)


class LoggingHandler(tornado.web.RequestHandler):
    """
        Base class for all request handlers. Logs the time, method, uri, and
        HTTP version to the console for each request.
    """

    def prepare(self):
        print("[{datetime}] {req.method} {req.uri} {req.version}".format(
            datetime=datetime.datetime.now().strftime("%m-%d-%Y %H:%M"),
            req=self.request))


class HomeHandler(LoggingHandler):
    """
        Renders the home page.
    """

    def get(self):
        self.render("home.html")


class GamesHandler(LoggingHandler):
    """
        Lists the available games to play.
    """

    def get(self):
        games = stratumgs.game.get_available_game_engines()
        self.render("games.html", games=games)


class ConfigureHandler(LoggingHandler):
    """
        Displays a configuration screen to configure a new game.
    """

    def get(self, game):
        players = stratumgs.client.server.get_available_client_names_for_game(game)
        config = stratumgs.game.get_game_configuration(game)
        self.render("configure.html", players=players, game_name=game, config=config)


class StartHandler(LoggingHandler):
    """
        Starts a new game.
    """

    def post(self, game):
        player_ids = self.get_arguments("players")
        game_id = stratumgs.game.init_game_engine(game, player_ids=player_ids)
        self.redirect("/games/tictactoe/view/{}".format(game_id))


class ViewHandler(LoggingHandler):
    """
        Displays a game.
    """

    def get(self, game, gid):
        game_template = self.render_string("games/{}.html".format(game))
        self.render("view.html", game_name=game, game_template=game_template)


class ViewSocketHandler(tornado.websocket.WebSocketHandler):
    """
        Handles websocket connections for viewing games.
    """

    def open(self, game, game_id):
        self.is_open = True
        stratumgs.game.get_game_runner(int(game_id)).add_view(self)

    def on_close(self):
        self.is_open = False

    def on_message(self, message):
        pass


class MatchesHandler(LoggingHandler):
    """
        Displays current matches.
    """

    def get(self):
        all_matches = stratumgs.game.get_current_games()
        active_matches = []
        inactive_matches = []
        for match in all_matches:
            if match[1].is_running:
                active_matches.append(match)
            else:
                inactive_matches.append(match)
        self.render("matches.html",
                    active_matches=active_matches,
                    inactive_matches=inactive_matches)


class PlayersHandler(LoggingHandler):
    """
        List currently connected players.
    """

    def get(self):
        players = stratumgs.client.server.get_connected_clients()
        self.render("players.html", players=players)



"""
.. module stratumgs.game.engine.client

This module contains the code that represents a client in the game engine.
"""

import json
import os
import socket


def init_engine_client(connection_info):
    """
        Initialize an engine client from the given endpoints. Chooses whether to
        create a pipe or socket based implementation, and returns the
        instantiated client.

        :param connection_info: The endpoints for the client connection.
        :returns: The engine client.
    """

    if os.name == "posix":
        return PipeEngineClient(connection_info)
    else:
        return SocketEngineClient(connection_info)


class PipeEngineClient(object):
    """
        An engine client implementation using pipes.

        :param file_descriptors: The read and write file descriptors for the
                                 pipes.
        :type file_descriptors: tuple(int, int)
    """

    def __init__(self, file_descriptors):
        self._read_pipe = None
        self._write_pipe = None
        if file_descriptors[0] is not None:
            self._read_pipe = open(file_descriptors[0], "rb")
        if file_descriptors[1] is not None:
            self._write_pipe = open(file_descriptors[1], "wb", buffering=0)

    def write(self, message):
        """
            Write a message to the client.

            :param message: The JSON-encodable message to write.
        """

        self._write_pipe.write("{}\n".format(json.dumps(message)).encode())

    def read(self):
        """
            Read a message from the client.

            :returns: A JSON-decoded object.
        """

        return json.loads(self._read_pipe.readline().decode().strip())

    def close(self, write_close=True):
        """
            Close the relevant connections.

            :param write_close: Whether or not to write the close message.
            :type write_close: boolean
        """

        if write_close:
            self.write({"type": "close"})
        if self._write_pipe:
            self._write_pipe.close()
        if self._read_pipe:
            self._read_pipe.close()


class SocketEngineClient(object):
    """
        An engine client implementation using sockets.

        :param port: The port to connect to.
        :type port: int
    """

    def __init__(self, port):
        self._socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM, 0)
        self._socket.connect(('127.0.0.1', port))
        self._socket_read_file = self._socket.makefile("rb", 0)
        self._socket_write_file = self._socket.makefile("wb", 0)

    def write(self, message):
        """
            Write a message to the client.

            :param message: The JSON-encodable message to write.
        """

        self._socket_write_file.write("{}\n".format(json.dumps(message)).encode())

    def read(self):
        """
            Read a message from the client.

            :returns: A JSON-decoded object.
        """

        return json.loads(self._socket_read_file.readline().decode().strip())

    def close(self, write_close=True):
        """
            Close the relevant connections.

            :param write_close: Whether or not to write the close message.
            :type write_close: boolean
        """

        if write_close:
            self.write({"type": "close"})
        self._socket_read_file.close()
        self._socket_write_file.close()
        self._socket.close()



"""
.. module stratumgs.game.runner

Classes and helpers in this module create and manage engine runners, which
manage the game engines which run as background processes.
"""

import json
import multiprocessing
import os

import tornado.ioloop

import tornado.iostream
import stratumgs.client.proxy

import stratumgs.game


def init_engine_runner(game_id, engine, engine_name, players):
    """
        Initialize a new engine runner. Since engine runners can be implemented
        using either pipes or sockets, depending on platform availability, this
        method simplifies creation of new runners.

        :param game_id: The ID of the game being created.
        :type game_id: int
        :param engine: The engine class to initialize in a background process.
        :type engine: ``stratumgs.game.engine.BaseEngine``
        :param engine_name: The name identifier of the engine to initialize.
        :type engine_name: string
        :param players: The players in the game.
        :type players: list(player endpoints)
        :returns: The created engine runner.
    """

    if os.name == "posix":
        return PipeEngineRunner(game_id, engine, engine_name, players)
    else:
        return SocketEngineRunner(game_id, engine, engine_name, players)


def _start_process(engine_constructor, player_endpoints, view_connection):
    """
        The target function for the background process. Intializes and starts
        the specified engine.

        :param engine_constructor: The engine class to initialize.
        :type engine_constructor: ``stratumgs.game.engine.BaseEngine``
        :param player_endpoints: The endpoints of the players in the game.
        :type player_endpoints: list(player endpoints)
        :param view_connection: The endpoints for the view connection.
    """

    engine = engine_constructor(players=player_endpoints, view_connection=view_connection)
    engine.run()


class BaseEngineRunner(object):
    """
        Most of the logic for running a game engine is contained in this class.
        It is extended by classes that implement the communication with the
        view, depending on whether pipes or sockets are in use.

        :param game_id: The ID of the game being created.
        :type game_id: int
        :param engine_constructor: The engine class to initialize in a background process.
        :type engine_constructor: ``stratumgs.game.engine.BaseEngine``
        :param engine_name: The name identifier of the engine to initialize.
        :type engine_name: string
        :param players: The players in the game.
        :type players: list(player endpoints)
    """

    def __init__(self, game_id, engine_constructor, engine_name, players):
        self._last_state = None
        self._connected_views = []

        engine_config = stratumgs.game.get_game_configuration(engine_name)
        self.engine_name = engine_name
        self.engine_display_name = engine_config["display_name"]
        self.is_running = True
        self.players = players

        view_connection = self.init_view_connection()

        player_endpoints = [player.create_endpoints_for_game(game_id) for player in players]
        engine_process = multiprocessing.Process(
            target=_start_process,
            args=(engine_constructor, player_endpoints, view_connection))
        engine_process.start()

        self.read_from_view_connection(b"\n", self._on_receive_state)

    def _on_receive_state(self, state):
        """
            Callback that is called when a game state is received over the view
            connection.

            :param state: The state that was sent from the engine.
            :type state: A JSON encoded string of the state.
        """

        obj = json.loads(state.decode().strip())
        if obj["type"] == "close":
            self.close_view_connection()
            self.is_running = False
            return
        self._last_state = state
        for view in self._connected_views[:]:
            if not view.is_open:
                self._connected_views.remove(view)
                continue
            view.write_message(state)
        self.read_from_view_connection(b"\n", self._on_receive_state)

    def add_view(self, view):
        """
            Add a view to the list of connected views.

            :param view: The view to add.
            :type view: :class:``tornado.websocket.WebSocketHandler``
        """

        if self._last_state:
            view.write_message(self._last_state)
        self._connected_views.append(view)


class PipeEngineRunner(BaseEngineRunner):
    """
        An implementation of an engine runner that uses pipes.
    """

    def init_view_connection(self):
        """
            Initializes the view connection using pipes.

            :returns: A tuple containing a ``None`` read endpoint and the write
                      endpoint of the pipe.
        """

        read_pipe, write_pipe = os.pipe()
        self._view_read_pipe = tornado.iostream.PipeIOStream(read_pipe)
        return (None, write_pipe)

    def read_from_view_connection(self, delimiter, callback):
        """
            Read from the view connection.

            :param delimiter: The delimeter to read until.
            :type delimiter: :class:``bytes``
            :param callback: The callback to call when the delimiter is reached.
            :type callbac: function
        """

        self._view_read_pipe.read_until(delimiter, callback)

    def close_view_connection(self):
        """
            Close the relevant connections.
        """

        self._view_read_pipe.close()


class SocketEngineRunner(BaseEngineRunner):
    """
        An implementation of an engine runner that uses sockets.
    """

    def init_view_connection(self):
        """
            Initializes the view connection using pipes.

            :returns: A tuple containing a ``None`` read endpoint and the write
                      endpoint of the pipe.
        """

        self.connector_server = stratumgs.client.proxy.SingleClientServer()
        return self.connector_server.get_port()

    def read_from_view_connection(self, delimiter, callback):
        """
            Read from the view connection.

            :param delimiter: The delimeter to read until.
            :type delimiter: :class:``bytes``
            :param callback: The callback to call when the delimiter is reached.
            :type callbac: function
        """

        tornado.ioloop.IOLoop.current().add_future(
            self.connector_server.get_stream(),
            lambda s: s.result().read_until(delimiter, callback))

    def close_view_connection(self):
        """
            Close the relevant connections.
        """

        tornado.ioloop.IOLoop.current().add_future(
            self.connector_server.get_stream(),
            lambda s: s.result().close())



import tornado.ioloop

import stratumgs.client.server
import stratumgs.config
import stratumgs.web


version = "0.1.1"


def main():
    debug = stratumgs.config.get("global", "debug")
    web_host = stratumgs.config.get("web_server", "host")
    web_port = stratumgs.config.get("web_server", "port")
    client_host = stratumgs.config.get("client_server", "host")
    client_port = stratumgs.config.get("client_server", "port")
    stratumgs.client.server.init(client_host, client_port)
    stratumgs.web.init(web_host, web_port, debug)
    tornado.ioloop.IOLoop.current().start()



"""
.. module stratumgs.client.server

The client server for stratumgs.
"""

import json

import tornado.ioloop
import tornado.iostream
import tornado.concurrent
import tornado.tcpserver

import stratumgs.client.proxy

_CONNECTED_CLIENTS = {}


def init(host, port):
    """
        Initialize the client server.

        :param host: The host to bind to.
        :type host: str
        :param port: The client server port.
        :type port: int
    """

    client_server = ClientProxyServer()
    client_server.listen(port, address=host)


def get_available_client_names_for_game(game):
    """
        Get a list of the names of the clients currently connected to the server
        with available game slots.

        :param game: The game that the clients need to support.
        :type game: string
    """

    return sorted(c.name for c in _CONNECTED_CLIENTS.values() if
                  c.is_available() and game in c.supported_games)


def get_connected_clients():
    """
        Get a list of all currently connected clients.
    """

    return sorted(_CONNECTED_CLIENTS.values(), key=lambda c: c.name)


def get_connected_client(client_name):
    """
        Get a connected client by name.

        :param client_name: The name of the client.
        :type client_name: string
    """

    return _CONNECTED_CLIENTS[client_name]


class StreamProxy(object):
    """
        A proxy for :class:`tornado.iostream.IOStream`. It only provides some
        methods of ``IOStream``, as follows: ``write``, ``read_until``,
        ``set_close_callback``, ``close``.

        :param stream: The stream to proxy.
        :type stream: :class:`tornado.iostream.IOStream`
    """

    def __init__(self, stream):
        self._stream = stream
        self._close_callback = None

    def write(self, message):
        self._stream.write(message)

    def read_until(self, delimeter, callback):
        self._stream.read_until(delimeter, callback)

    def set_close_callback(self, callback):
        self._close_callback = callback

    def close(self):
        self._close_callback()


class ClientProxyServer(tornado.tcpserver.TCPServer):
    """
        Listens for new clients and connects them.
    """

    _NAMELESS_CLIENT_NUMBER = 1

    @classmethod
    def _negotiate_name(cls, name):
        """
            Private helper method used to determine a unique client name,
            possibly given a name as a starting point.

            If given name is unique among the currently connected client, then
            this method returns that. If it isn't, appends a number to it to
            make it unique. If the name provided is ``None``, creates a new name
            of the form ``name-#`` where ``#`` is a number that makes the name
            unique.

            :param name: The name.
            :type name: string
            :returns: A unique name for the client.
        """

        if name is None:
            name = "client-{}".format(cls._NAMELESS_CLIENT_NUMBER)
            cls._NAMELESS_CLIENT_NUMBER += 1
        elif name in _CONNECTED_CLIENTS:
            n = 1
            while True:
                possible_name = "{}-{}".format(name, n)
                if possible_name not in _CONNECTED_CLIENTS:
                    name = possible_name
                    break
        return name

    def handle_stream(self, stream, address):

        def new_client(connect_message):
            connect_message = json.loads(connect_message.decode().strip())

            if connect_message["type"] != "connect":
                print("Invalid message type from client {}".format(address))
                return

            name = self._negotiate_name(connect_message["name"])
            try:
                max_games = int(connect_message["max_games"])
            except ValueError:
                print("Invalid max_games parameter from client {}".format(address))
                return
            supported_games = connect_message["supported_games"]

            stream.write("{}\n".format(json.dumps({
                "type": "name",
                "name": name
            })).encode())

            stream_proxy = StreamProxy(stream)

            def stream_closed():
                print("Client {} disconnected.".format(name))
                del _CONNECTED_CLIENTS[name]
                stream_proxy.close()

            stream.set_close_callback(stream_closed)

            _CONNECTED_CLIENTS[name] = stratumgs.client.proxy.ClientProxy(
                name, supported_games, max_games, stream_proxy)

            print("Client {} connected.".format(name))

        stream.read_until(b"\n", new_client)



"""
.. module stratumgs.game

Contains the functionality to create new games and query information about
existing games.
"""

import stratumgs.game.runner
import stratumgs.client.server

from .games import tictactoe


_GAME_ENGINES = {
    "tictactoe": tictactoe
}

_CREATED_GAME_ID = 0
_CREATED_GAMES = {}


def get_available_game_engines():
    """
        Get a list of the available game engines.

        :returns: A list of tuples of the form ``(game_id, game_configuration)``.
    """

    return [(game_key, get_game_configuration(game_key))
            for game_key in sorted(_GAME_ENGINES.keys())]


def init_game_engine(engine_name, player_ids=[]):
    """
        Initialize a new game engine.

        :param engine_name: The name of the engine to initialize.
        :type engine_name: string
        :param player_ids: A list of the player IDs in the game.
        :type player_ids: list(string)
        :returns: The ID of the newly created game.
    """

    global _CREATED_GAME_ID
    game_id = _CREATED_GAME_ID
    _CREATED_GAME_ID += 1
    players = [stratumgs.client.server.get_connected_client(pid) for pid in player_ids]
    engine_runner = stratumgs.game.runner.init_engine_runner(
        game_id, _GAME_ENGINES[engine_name].Engine, engine_name, players)
    _CREATED_GAMES[game_id] = engine_runner
    return game_id


def get_game_configuration(game_name):
    """
        Get the configuration for the specified game.

        :param game_name: The name of the game engine to get the configuration for.
        :type game_name: string
        :returns: The game configuration.
    """

    return _GAME_ENGINES[game_name].CONFIG


def get_current_games():
    """
        Get a list of all currently being played games.

        :returns: A list of all current games, in unspecified order.
    """

    return _CREATED_GAMES.items()


def get_game_runner(game_id):
    """
        Get a specific game runner by ID.

        :param game_id: The game ID to retrieve the runner for.
        :type game_id: int
        :returns: The runner.
    """

    return _CREATED_GAMES[game_id]



from setuptools import setup, find_packages

import stratumgs


def readme():
    with open("README.rst") as f:
        return f.read()


setup(name="stratumgs",
      version=stratumgs.version,
      description="A turn based game engine designed to pit autonomous players "
                  "against each other.",
      long_description=readme(),
      classifiers=[
            "Development Status :: 3 - Alpha",
            "Environment :: Console",
            "Environment :: Web Environment",
            "License :: OSI Approved :: MIT License",
            "Programming Language :: Python :: 3",
            "Topic :: Games/Entertainment :: Board Games",
            "Topic :: Games/Entertainment :: Puzzle Games",
            "Topic :: Games/Entertainment :: Turn Based Strategy"
      ],
      entry_points={
            "console_scripts": ["stratumgs=stratumgs:main"]
      },
      keywords=["stratumgs", "stratum", "game", "server", "turn", "based",
                "board", "ai", "autonomous", "tictactoe"],
      url="https://stratumgs.org",
      author="David Korhumel",
      author_email="dpk2442@gmail.com",
      license="MIT",
      packages=find_packages(),
      install_requires=["tornado"],
      include_package_data=True,
      zip_safe=False)



"""
.. module stratumgs.config

Loads the configuration and provides a method for accessing it.
"""

import configparser
import os


# The default value and type of each configuration parameter
_CONFIG_VALUES = {
    "global": {
        "debug": (bool, True)
    },
    "web_server": {
        "host": (str, ""),
        "port": (int, 8888)
    },
    "client_server": {
        "host": (str, ""),
        "port": (int, 8889)
    }
}

# The configuration object itself
_CONFIG = configparser.ConfigParser()

# Look for stratumgs.conf in project root.
# If that doesn't exist, look for /etc/stratumgs.conf.
path = os.path.abspath(
    os.path.join(os.path.dirname(__file__), "..", "stratumgs.conf"))
if os.path.exists(path):
    _CONFIG.read(path)
else:
    path = "/etc/stratumgs.conf"
    if os.path.exists(path):
        _CONFIG.read(path)


def get(section, option):
    """
        Get a value from the configuration.

        :param section: The section in the config the value is in.
        :type section: str
        :param option: The option within the section.
        :type option: str
        :returns: The value for the configuration parameter, or the default
                  value if no explicit value was set.
    """

    if section in _CONFIG_VALUES:
        if option in _CONFIG_VALUES[section]:
            type_fn, default = _CONFIG_VALUES[section][option]
            return type_fn(_CONFIG.get(section, option, fallback=default))
    return None



"""
.. module stratumgs.client.proxy

Proxies clients for the game engines.
"""

import json
import os

import tornado.concurrent
import tornado.ioloop
import tornado.iostream
import tornado.netutil
import tornado.tcpserver

import stratumgs.game


def _make_pipe_pair():
    """
        Make two pipes, and return tuples of the read and write pairs.

        :returns: Read/write pairs for the two pipes.
    """

    r1, w1 = os.pipe()
    r2, w2 = os.pipe()
    return (r1, w2), (r2, w1)


class ClientProxy(object):
    """
        Proxies the raw client stream for the game engine.

        :param name: The client name.
        :type name: string
        :param max_games: The maximum number of simultaneous games the client can support.
        :type max_games: int
        :param stream: The raw client stream.
        :type stream: :class:`tornado.iostream.IOStream`
    """

    def __init__(self, name, supported_games, max_games, stream):
        self.name = name
        self.supported_games = supported_games
        self.max_games = max_games
        self.games_available = max_games
        self.stream = stream
        self.helpers = {}

        self.supported_games_display = []
        for game in supported_games:
            try:
                self.supported_games_display.append(
                    stratumgs.game.get_game_configuration(game)["display_name"])
            except:
                pass
        self.supported_games_display.sort()

        def stream_closed():
            for helper in self.helpers.values():
                helper.write_to_engine("{}\n".format(json.dumps({
                    "type": "close"
                })).encode())
                helper.close_engine_connection_endpoints()

        def message_from_client(msg):
            obj = json.loads(msg.decode().strip())
            if obj["type"] == "close":
                self.stream.close()
                return
            if obj["game_id"] in self.helpers:
                self.helpers[obj["game_id"]].write_to_engine(msg)
            self.stream.read_until(b"\n", message_from_client)

        self.stream.set_close_callback(stream_closed)
        self.stream.read_until(b"\n", message_from_client)

    def is_available(self):
        """
            Determine whether the client has game slots available.

            :returns: Whether the client has game slots available.
        """

        return self.games_available > 0

    def create_endpoints_for_game(self, game_id):
        """
            Create a set of endpoints to be used by a game engine.

            :param game_id: The id of the game being created.
            :type game_id: int
            :returns: The endpoints for the game engine to use to connect.
        """

        self.games_available -= 1

        if os.name == "posix":
            helper = PipeClientProxyHelper()
        else:
            helper = SocketClientProxyHelper()

        self.helpers[game_id] = helper

        def message_from_engine(msg):
            # decode object and inject game id
            obj = json.loads(msg.decode().strip())
            obj["game_id"] = game_id

            # encode and write object to client
            self.stream.write("{}\n".format(json.dumps(obj)).encode())

            # if engine closing connection, clean up endpoints,
            # make game available, and return
            if obj["type"] == "close":
                helper.close_engine_connection_endpoints()
                del self.helpers[game_id]
                self.games_available += 1
                return

            # read from the engine stream again
            helper.read_from_engine(b"\n", message_from_engine)

        endpoints = helper.init_engine_connection_endpoints()

        helper.read_from_engine(b"\n", message_from_engine)

        self.stream.write("{}\n".format(json.dumps({
            "type": "start",
            "game_id": game_id
        })).encode())

        return endpoints


class PipeClientProxyHelper():
    """
        A helper object to manage connection endpoints. This helpers uses pipes
        to handle communication between the client and engine.
    """

    def init_engine_connection_endpoints(self):
        """
            Initialize the endpoints for the engine.

            :returns: The endpoints.
        """

        client_end, engine_end = _make_pipe_pair()
        self.to_engine = tornado.iostream.PipeIOStream(client_end[1])
        self.from_engine = tornado.iostream.PipeIOStream(client_end[0])
        return engine_end

    def close_engine_connection_endpoints(self):
        """
            Close the endpoints to the engine.
        """

        self.to_engine.close()
        self.from_engine.close()

    def write_to_engine(self, msg):
        """
            Write a message to the engine.

            :param msg: The message to write.
            :type msg: :class:`bytes`
        """

        if not self.to_engine.closed():
            self.to_engine.write(msg)

    def read_from_engine(self, delimeter, callback):
        """
            Read from the engine until the given delimiter is reached. The read
            message will be passed to the callback function.

            :param delimiter: The delimiter to read until.
            :type delimiter: :class:`bytes`
            :param callback: The callback to call.
            :type callback: function
        """

        self.from_engine.read_until(delimeter, callback)


class SocketClientProxyHelper():
    """
        A helper object to manage connection endpoints. This helpers uses
        sockets to handle communication between the client and engine.
    """

    def init_engine_connection_endpoints(self):
        """
            Initialize the endpoints for the engine.

            :returns: The endpoints.
        """

        self.connector_server = SingleClientServer()
        return self.connector_server.get_port()

    def close_engine_connection_endpoints(self):
        """
            Close the endpoints to the engine.
        """

        tornado.ioloop.IOLoop.current().add_future(
            self.connector_server.get_stream(),
            lambda s: s.result().close())

    def write_to_engine(self, msg):
        """
            Write a message to the engine.

            :param msg: The message to write.
            :type msg: :class:`bytes`
        """

        def cb(s):
            if not s.result().closed():
                s.result().write(msg)
        tornado.ioloop.IOLoop.current().add_future(
            self.connector_server.get_stream(), cb)

    def read_from_engine(self, delimeter, callback):
        """
            Read from the engine until the given delimiter is reached. The read
            message will be passed to the callback function.

            :param delimiter: The delimiter to read until.
            :type delimiter: :class:`bytes`
            :param callback: The callback to call.
            :type callback: function
        """

        tornado.ioloop.IOLoop.current().add_future(
            self.connector_server.get_stream(),
            lambda s: s.result().read_until(delimeter, callback))


class SingleClientServer(tornado.tcpserver.TCPServer):
    """
        A TCP server that runs on an available port, allows a single connection,
        and stores the single connection.
    """

    def __init__(self):
        super(SingleClientServer, self).__init__()
        self._stream_future = tornado.concurrent.Future()
        sockets = tornado.netutil.bind_sockets(0, '127.0.0.1')
        self._port = sockets[0].getsockname()[1]
        self.add_sockets(sockets)
        print(sockets)

    def get_port(self):
        """
            Get the port of the server.

            :returns: The server port.
        """

        return self._port

    def handle_stream(self, stream, address):
        if not self._stream_future.done():
            self._stream_future.set_result(stream)
        self.io_loop.add_callback(lambda: self.stop())

    def get_stream(self):
        """
            Get a future for the stream that connects.

            :returns: A :class:`tornado.concurrent.Future` that will hold the
                      connected stream.
        """

        return self._stream_future



"""
.. module stratumgs.game.engine

This module contains the base engine that all game engines extend, as well as
code for interacting with clients.
"""

# flake8:noqa

from .engine import BaseEngine



import json
import sys

from .client import init_engine_client


class BaseEngine(object):
    """
        Contains the base code for all game engines. Engines must extend this
        class, and need to implement ``is_game_over``, ``get_state``, and
        ``play_turn``.

        :param players: The list of player endpoints.
        :type players: list(player endpoints)
        :param view_connection: The view connection endpoints.
    """

    def __init__(self, players=[], view_connection=None):
        self.num_players = len(players)
        self._player_clients = [init_engine_client(p) for p in players]
        self._view_client = init_engine_client(view_connection)

    def _send_state(self):
        """
            Send the current state of the game to the view client.
        """

        msg = {
            "type": "message",
            "payload": json.dumps(self.get_state())
        }
        for p in self._player_clients:
            p.write(msg)
        self._view_client.write(msg)

    def run(self):
        """
            Start the main game loop.
        """

        self._send_state()
        while not self.is_game_over():
            self.play_turn()
            self._send_state()
        self._send_state()
        for p in self._player_clients:
            p.close()
        self._view_client.close()

    def send_message_to_player(self, player_id, message):
        """
            Send the given message to the given player ID. The message gets
            encoded and sent to the client.

            :param player_id: The ID of the player to send a message to.
            :type player_id: int
            :param message: The message to send.
            :type message: string
        """

        self._player_clients[player_id].write({
            "type": "message",
            "payload": json.dumps(message)
        })

    def receive_message_from_player(self, player_id):
        """
            A method that blocks until a message is received from the given
            player.

            :param player_id: The ID of the player to receive a message from.
            :type player_id: int
            :returns: The message received from the player.
        """

        obj = self._player_clients[player_id].read()
        if obj["type"] == "close":
            print("Player id {} disconnected.".format(player_id))
            for p in self._player_clients:
                p.close()
            self._view_client.close()
            sys.exit(1)
        return json.loads(obj["payload"])

    def is_game_over(self):
        """
            Must be implemented by the game engine. Checks if the game is over.

            :returns: Tue if the game is over, false otherwise.
        """

        raise NotImplementedError

    def get_state(self):
        """
            Must be implemented by the game engine. Get the current state of the
            game.

            :returns: The state of the game.
        """

        raise NotImplementedError

    def play_turn(self):
        """
            Must be implemented by the game engine. Plays a turn of the game.
        """

        raise NotImplementedError



"""
.. module stratumgs.game.games.tictactoe

A TicTacToe game engine.
"""

from ..engine import BaseEngine

CONFIG = {
    "display_name": "TicTacToe",
    "description": "A game of TicTacToe.",
    "num_players": 2,
    "player_names": ["X", "O"]
}


def _get_first_if_all_equal(lst):
    """
        Get the first element of a list if all the elements of the list are
        equivalent.

        :param lst: The list of elements.
        :type lst: list
    """

    first = lst[0]
    for el in lst[1:]:
        if el != first:
            return None
    return first


class Engine(BaseEngine):
    """
        The engine class for TicTacToe.

        :param players: The players for the game.
        :type players: list(player endpoints)
        :param view_connection: The view connection endpoints.
    """

    def __init__(self, players=[], view_connection=None):
        super(Engine, self).__init__(players=players, view_connection=view_connection)
        self._board = [[None for _ in range(3)] for _ in range(3)]
        self._winner = None
        self._x_turn = True

    def is_game_over(self):
        """
            Check if the game is over. The game is over when a player gets 3 in
            a row, or when all spaces on the board are full.

            :returns: True if the game is over, false otherwise.
        """

        for i in range(3):
            row_val = _get_first_if_all_equal(self._board[i])
            col_val = _get_first_if_all_equal(tuple(self._board[j][i] for j in range(3)))
            if row_val == "X" or col_val == "X":
                self._winner = "X"
                return True
            elif row_val == "O" or col_val == "O":
                self._winner = "O"
                return True
        diag1_val = _get_first_if_all_equal(tuple(self._board[i][i] for i in range(3)))
        diag2_val = _get_first_if_all_equal(tuple(self._board[2-i][i] for i in range(3)))
        if diag1_val == "X" or diag2_val == "X":
            self._winner = "X"
            return True
        elif diag1_val == "O" or diag2_val == "O":
            self._winner = "O"
            return True
        return not any(cell is None for row in self._board for cell in row)

    def get_state(self):
        """
            Get the current state object for the game. The state of the game
            includes the contents of the board, and the winner of the game.
            Before the game is over or when the outcome is a draw, the winner is
            ``None``.

            :returns: The state.
        """

        return {
            "type": "state",
            "board": self._board,
            "winner": self._winner
        }

    def play_turn(self):
        """
            Play a turn of the game. The current player is prompted to make a
            move. If the move is invalid, the client is asked to repeat the
            turn. Once the client has input a valid move, the move is recorded,
            and the turn ends.
        """

        cur_player_id, cur_player_letter = (0, "X") if self._x_turn else (1, "O")
        self.send_message_to_player(cur_player_id, {"type": "turn"})
        while True:
            move = self.receive_message_from_player(cur_player_id)
            row, col = move["row"], move["column"]
            error = None
            if row > 2 or col > 2:
                error = "out-of-bounds"
            elif self._board[row][col] is not None:
                error = "space-not-empty"
            if error is None:
                break
            self.send_message_to_player(cur_player_id, {
                "type": "repeat-turn",
                "error": error,
                "last-move": move
            })
        self._board[move["row"]][move["column"]] = cur_player_letter
        self._x_turn = not self._x_turn
